#!/bin/ash
# Fatdog Redshift GUI
# Copyright (C) James Budiono 2018, 2020
#
# License: GNU GPL Version 3 or later
#
# Note: the bits in Global Inits, which is part of GTK-server bash init code
#       is (C) Peter van Eerten. Everything else is (C) James Budiono
#

# std localisation stanza
export TEXTDOMAIN=fatdog
. gettext.sh
# performance tweak - use "C" if there is no localisation
! [ -e $TEXTDOMAINDIR/${LANG%.*}/LC_MESSAGES/$TEXTDOMAIN.mo ] &&
! [ -e $TEXTDOMAINDIR/${LANG%_*}/LC_MESSAGES/$TEXTDOMAIN.mo ] && LANG=C

### configuration
APPTITLE="$(gettext 'Redshift GUI')"
MAP_ZOOM_LEVEL=15 # starting zoom level
MAP_WIDTH=400
MAP_HEIGHT=400
MAP_TYPE=mapnik
CONFFILE=~/.config/redshift.conf
MAP_SOURCE=OSM1 # or OSM2, or WIKIMEDIA, etc

### runtime variables 
GDK_TYPE_PIXBUF= MAIN_WINDOW= NOTEBOOK=
TMPDIR=/tmp/redshiftgui.$$
PREV_ADDR=



################# START OF MAP-FETCH library ###########################
#!/bin/dash

### list of free, open tileservers (no registration)
# from here: https://wiki.openstreetmap.org/wiki/Tile_servers
#
OSM1='https://a.tile.openstreetmap.org/${z}/${x}/${y}.png'
OSM2='https://b.tile.openstreetmap.org/${z}/${x}/${y}.png'
OSM3='https://c.tile.openstreetmap.org/${z}/${x}/${y}.png'

OSMDE1='https://a.tile.openstreetmap.de/${z}/${x}/${y}.png'
OSMDE2='https://b.tile.openstreetmap.de/${z}/${x}/${y}.png'
OSMDE3='https://c.tile.openstreetmap.de/${z}/${x}/${y}.png'

WIKIMEDIA='https://maps.wikimedia.org/osm-intl/${z}/${x}/${y}.png'

HUMANITARIAN1='http://a.tile.openstreetmap.fr/hot/${z}/${x}/${y}.png'
HUMANITARIAN2='http://b.tile.openstreetmap.fr/hot/${z}/${x}/${y}.png'

HIKEBIKE='https://tiles.wmflabs.org/hikebike/${z}/${x}/${y}.png'
HILLSHADING='http://tiles.wmflabs.org/hillshading/${z}/${x}/${y}.png'
MAPNIKBW='https://tiles.wmflabs.org/bw-mapnik/${z}/${x}/${y}.png'
MAPNIK_NO_LABEL='https://tiles.wmflabs.org/osm-no-labels/${z}/${x}/${y}.png'

STAMEN_BW='http://a.tile.stamen.com/toner/${z}/${x}/${y}.png'
STAMEN_WC='http://c.tile.stamen.com/watercolor/${z}/${x}/${y}.jpg'

OPENKARTE='http://tile.memomaps.de/tilegen/${z}/${x}/${y}.png'

OPENTOPO='https://{a|b|c}.tile.opentopomap.org/{z}/{x}/{y}.png'

### helper functions

# $1-x, $1-y, $3-zoom, $4-tileserver
apply_values() {
	local svr="$4" x="$1" y="$2" z="$3"
	echo "$svr" | sed "s|\${x}|$x|;s|\${y}|$y|;s|\${z}|$z|"
}

# $1-x, $1-y, $3-zoom, $4-tileserver, $5-destination (blank=stdout)
get_map_xy() {
	if [ -z "$5" ]; then
		curl "$(apply_values $1 $2 $3 "$4")"
	else
		curl "$(apply_values $1 $2 $3 "$4")" > "$5"
	fi
}

# $1-lat, $2-long, $3-zoom, $4-tileserver, $5-destination (blank=stdout)
get_map() {
	local z=$3
	local y=$(lat2ytile $1 $z) x=$(long2xtile $2 $z)
	echo $x $y $z > /dev/stderr
	get_map_xy $x $y $z "$4" "$5"
}


### IMPORTED STUFF #####################################################
# source: https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Bourne_shell_with_Awk
# tidied and reformatted and commented#

# $1-xtile, $2-zoom level, output: longitude
xtile2long()
{
	xtile=$1
	zoom=$2
	echo "${xtile} ${zoom}" | awk '{printf("%.9f", $1 / 2.0^$2 * 360.0 - 180)}'
} 

# $1-longitude, $2-zoom level, output: xtile
long2xtile()  
{ 
	long=$1
	zoom=$2
	echo "${long} ${zoom}" |
	awk '
	{ xtile = ($1 + 180.0) / 360 * 2.0^$2; 
	xtile+=xtile<0?-0.5:0.5;
	printf("%d", xtile ) }'
}

# $1-ytile, $2-zoom level, $3-(boolean) ytile is specified in "tms" format
# output: latitude
ytile2lat()
{
	ytile=$1;
	zoom=$2;
	tms=$3;
	if [ ! -z "${tms}" ]
	then
		# from tms_numbering into osm_numbering
		ytile=`echo "${ytile}" ${zoom} | awk '{printf("%d\n",((2.0^$2)-1)-$1)}'`;
	fi
	lat=$(echo "${ytile} ${zoom}" |
		awk -v PI=3.14159265358979323846 '
		{ 
			num_tiles = PI - 2.0 * PI * $1 / 2.0^$2;
			printf("%.9f", 180.0 / PI * atan2(0.5 * (exp(num_tiles) - exp(-num_tiles)),1));
		}');
	echo "${lat}";
}

# $1-latitude, $2-zoom level, $3-(boolean) ytile is specified in "tms" format
# output: ytile
lat2ytile() 
{ 
	lat=$1;
	zoom=$2;
	tms=$3;
	ytile=$(echo "${lat} ${zoom}" |
		awk -v PI=3.14159265358979323846 '{ 
		tan_x=sin($1 * PI / 180.0)/cos($1 * PI / 180.0);
		ytile = (1 - log(tan_x + 1/cos($1 * PI/ 180))/PI)/2 * 2.0^$2; 
		ytile+=ytile<0?-0.5:0.5;
		printf("%d", ytile )
		}');
	if [ ! -z "${tms}" ]
	then
		# from oms_numbering into tms_numbering
		ytile=`echo "${ytile}" ${zoom} | awk '{printf("%d\n",((2.0^$2)-1)-$1)}'`;
	fi
	echo "${ytile}";
}

# test cases we don't use
<< "EOF"
# ------------------------------------
# Sample of use: 
# Position Brandenburg Gate, Berlin
# ------------------------------------
LONG=13.37771496361961;
LAT=52.51628011262304;
ZOOM=17;
TILE_X=70406;
TILE_Y=42987; 
TILE_Y_TMS=88084;
TMS=""; # when NOT empty: tms format assumed

# ------------------------------------
# assume input/output of y is in oms-format:
LONG=$( xtile2long ${TILE_X} ${ZOOM} );
LAT=$( ytile2lat ${TILE_Y} ${ZOOM} ${TMS} );

# Result should be longitude[13.375854492] latitude[52.517892228]
TILE_X=$( long2xtile ${LONG} ${ZOOM} );
TILE_Y=$( lat2ytile ${LAT} ${ZOOM} ${TMS} );
# Result should be x[70406] y_oms[42987]
# ------------------------------------

# assume input/output of y is in tms-format:
TMS="tms";
TILE_Y_TMS=$( lat2ytile ${LAT} ${ZOOM} ${TMS} );
LAT_TMS=$( ytile2lat ${TILE_Y_TMS} ${ZOOM} ${TMS} );
echo "Result should be y_oms[${TILE_Y}] latitude[${LAT}] ; y_tms[${TILE_Y_TMS}] latitude_tms[${LAT_TMS}] "
# latitude and latitude_tms should have the same value ; y_oms and y_tms should have the given start values:
# Result should be y_oms[42987] latitude[52.517892228] ; y_tms[88084] latitude_tms[52.517892228]
# ------------------------------------
EOF
### END OF IMPORTED STUFF ##############################################

# tests
#curl "$(apply_values 63 42 7 "$OSM1")" | pngtopam | pamx
#get_map_xy 63 42 7 "$OSM1" | pngtopam | pamx
#get_map 51.5132 -0.1032 7 "$OSM1" | pngtopam | pamx

################# END OF MAP-FETCH library #############################



# tree model

#---------------------------------------------------------- Global Inits
### global variables for proper GTK-server
GTK= NULL="NULL"
PIPE=/tmp/gtk.bash.$$
trap 'stop-gtk-server -fifo $PIPE; rm -r $TMPDIR; exit' HUP TERM INT # stop GTK-server in case of an error

### start GTK-server
gtk-server -fifo=$PIPE &
while [ ! -p $PIPE ]; do sleep 1; continue; done

### GTK-server helper functions
# Assignment function
define() { $2 $3 $4 $5 $6 $7 $8 $9; eval $1=\'"$GTK"\'; }

# Communicate with GTK-server
gtk()
{
/bin/echo "$@" > $PIPE
read GTK < $PIPE
}

#---------------------------------------------------------- GUI realization

# add GTK+ API not defined in gtk-server.cfg
add_missing_gtk_api() {
	gtk gtk_server_define gdk_pixbuf_get_type NONE INT 0
	gtk gtk_server_define gtk_label_set_markup NONE NONE 2 WIDGET STRING
	
	# add GTK+ API not defined in gtk-server.cfg
	gtk gtk_server_define gtk_builder_new NONE WIDGET 0
	gtk gtk_server_define gtk_builder_add_from_file NONE INT 3 WIDGET STRING NULL
	gtk gtk_server_define gtk_builder_add_from_string NONE INT 4 WIDGET STRING INT NULL
	gtk gtk_server_define gtk_builder_get_object NONE WIDGET 2 WIDGET STRING
	gtk gtk_server_define gtk_builder_connect_signals NONE NONE 2 WIDGET NULL	
	
	gtk gtk_server_define gtk_tree_view_get_cursor NONE NONE 3 WIDGET PTR_LONG NULL
	
	# hack for getting unlimited string length and leak-free operation with gtk_tree_get_model
	# see below
	gtk gtk_server_define strcat NONE STRING 2 POINTER STRING
	gtk gtk_server_define memcpy NONE NONE 3 PTR_LONG POINTER INT			
}

### helper - leak-free, unlimited string length from gtk_tree_model_get with unpatched gtk-server
# works with gtk-server 2.3.1 and 2.4.5
# $1-tree $2-iter $3-index $4-result name
tree_model_get_string() {
	local PSTR # pointer to string
	gtk gtk_server_define gtk_tree_model_get NONE NONE 5 WIDGET WIDGET INT PTR_LONG INT # ensure correct definition
	define PSTR gtk gtk_tree_model_get $1 $2 $3 NULL -1 # PSTR --> *buf	
	
	if [ $PSTR -ne 0 ]; then
	# with newer gtk-server 2.4.4 onwards you can use this
		#define $4 gtk gtk_server_string_from_pointer $PSTR 
		define $4 gtk strcat $PSTR \"\" # get string point by address --> **buf
		gtk free $PSTR # this was dynamically alloc-ed by gtk_tree_model_get, now free it
	else
		eval $4=\"\"
	fi
}


### generic helper
# $1-text
msg() {
	[ -z "$DISPLAY" ] && echo "$1" ||
	Xdialog --title "$APPTITLE" --infobox "$1" 0 0 10000
}

# $1-text, no-text => kill previous splash
splash() {
	if [ "$1" ]; then
		if [ "$DISPLAY" ]; then
			Xdialog --title "$APPTITLE" --no-buttons --infobox "$1" 0 0 10000000 &
			SPID=$!
		else
			echo "$1"
		fi
	else 
		[ "$SPID" ] && kill $SPID 2>/dev/null
		SPID=""
	fi
}

################### event handlers ###################
# LAT LON is the location of the map
load_map() {
	local LAT LON

	# get the coordinate
	define LAT gtk gtk_entry_get_text $LAT_ENTRY
	define LON gtk gtk_entry_get_text $LON_ENTRY
	[ -z "$LAT" -o -z "$LON" ] && return # sanity check
	
	# get zoom level
	define MAP_ZOOM_LEVEL gtk gtk_adjustment_get_value $MAP_SCALE
	MAP_ZOOM_LEVEL=$(printf "%.0f\n" $MAP_ZOOM_LEVEL)
	#echo $MAP_ZOOM_LEVEL

	# now downloadload the map image
	splash "Downloading map, please wait ..."
	get_map "$LAT" "$LON" "$MAP_ZOOM_LEVEL" "$(eval echo \$$MAP_SOURCE)" "$TMPDIR/map.png"	
	splash

	# show the map
	define IMG gtk gdk_pixbuf_new_from_file "$TMPDIR/map.png" NULL
	gtk gtk_image_set_from_pixbuf $MAP_IMAGE $IMG
	gtk g_object_unref $IMG
}

# returns PREV_ADDR
address_to_latlon() {
	local ADDR LATLONG IMG LAT LON

	# get the text, find the latlong
	define ADDR gtk gtk_entry_get_text $SEARCH_ENTRY
	#echo $ADDR

	if [ "$ADDR" ] && [ "$PREV_ADDR" != "$ADDR" ]; then
		splash "Finding geo-coordinate for the address, please wait ..."
		LATLONG="$(wget -qO - "https://nominatim.openstreetmap.org/search?format=xml&limit=1&q=$ADDR" | xml2 | grep -E '@lat|@lon')"
		splash
		if [ "$LATLONG" ]; then
			LAT=$(echo "$LATLONG" | sed '/@lat/!d;s/.*@lat=//')
			LON=$(echo "$LATLONG" | sed '/@lon/!d;s/.*@lon=//')
			#echo $LAT
			#echo $LON
			gtk gtk_entry_set_text $LAT_ENTRY $LAT
			gtk gtk_entry_set_text $LON_ENTRY $LON
			PREV_ADDR="$ADDR"			
		else
			msg "Cannot find location. Please specify the address in a different way."
			return
		fi
	fi
}

# 
load_settings() {
	local p LAT LON
	if [ -f $CONFFILE ]; then
		while read -r p; do
			case "$p" in
				lon=*) LON=${p#lon=} ;;
				lat=*) LAT=${p#lat=} ;;
			esac
		done < $CONFFILE

		[ -z "$LAT" -o -z "$LON" ] && return
		gtk gtk_entry_set_text $LAT_ENTRY $LAT
		gtk gtk_entry_set_text $LON_ENTRY $LON
	fi
}

# 
save_settings() {
	# get the coordinate
	define LAT gtk gtk_entry_get_text $LAT_ENTRY
	define LON gtk gtk_entry_get_text $LON_ENTRY
	[ -z "$LAT" -o -z "$LON" ] && return # sanity check
	
	if ! [ -f $CONFFILE ]; then
		# if doesn't exist, create
		> $CONFFILE cat << EOF
[manual]
lat=$LAT
lon=$LON
EOF
	else
		# otherwise, sed
		sed -i -e "/\[manual]/,$ {s/lat=.*/lat=$LAT/;s/lon=.*/lon=$LON/}" $CONFFILE
	fi
	msg "$(gettext "Settings saved.")"
}

################### main ###################
### gtk-server localisation stanza
gtk gtk_server_define setlocale NONE STRING 2 INT STRING
gtk gtk_server_define bindtextdomain NONE STRING 2 STRING STRING
gtk gtk_server_define textdomain NONE STRING 1 STRING
gtk setlocale 6 \"\" # 6 corresponds to LC_ALL
gtk bindtextdomain $TEXTDOMAIN $TEXTDOMAINDIR
gtk textdomain $TEXTDOMAIN


### init gui
add_missing_gtk_api
gtk gtk_init
define GDK_TYPE_PIXBUF gtk gdk_pixbuf_get_type

define BUILDER gtk gtk_builder_new
define RESULT gtk gtk_builder_add_from_file $BUILDER "${0}.xml"
gtk gtk_builder_connect_signals $BUILDER

define MAIN_WINDOW gtk gtk_builder_get_object $BUILDER main_window
define SAVE_BUTTON gtk gtk_builder_get_object $BUILDER save-button
define QUIT_BUTTON gtk gtk_builder_get_object $BUILDER quit-button
define SEARCH_BUTTON gtk gtk_builder_get_object $BUILDER search-button

define LAT_ENTRY gtk gtk_builder_get_object $BUILDER lat-entry
define LON_ENTRY gtk gtk_builder_get_object $BUILDER lon-entry
define SEARCH_ENTRY gtk gtk_builder_get_object $BUILDER search-entry

define MAP_IMAGE gtk gtk_builder_get_object $BUILDER map-image
define MAP_SCALE gtk gtk_builder_get_object $BUILDER map-scale-adjust
define MAP_RELOAD_BUTTON gtk gtk_builder_get_object $BUILDER map-reload

# connect signals
gtk gtk_server_connect $MAIN_WINDOW delete-event quit
gtk gtk_server_connect $SAVE_BUTTON clicked save
gtk gtk_server_connect $QUIT_BUTTON clicked quit
gtk gtk_server_connect $SEARCH_BUTTON clicked search
gtk gtk_server_connect $MAP_RELOAD_BUTTON clicked map-reload

# preparation
mkdir $TMPDIR
gtk gtk_adjustment_set_value $MAP_SCALE $MAP_ZOOM_LEVEL
load_settings
load_map

### gtk main loop
gtk gtk_widget_show_all $MAIN_WINDOW
while true; do 
	define EVENT gtk gtk_server_callback wait
	case $EVENT in
		quit)	break ;;
		save)	save_settings ;;
		map-reload) load_map ;;		
		search)
			address_to_latlon
			load_map
			;;
	esac
done
gtk gtk_server_exit 
rm -r $TMPDIR
